if select(2, UnitClass("player")) ~= "PRIEST" then return end

local name, addon = ...

_G[name] = addon

local pairs, ipairs = pairs, ipairs
local select = select
local format = format
local sqrt = sqrt
local gsub = gsub
local CreateFrame = CreateFrame
local UnitBuff = UnitBuff
local UnitPower = UnitPower

local shadowSpec = SPEC_PRIEST_SHADOW
local powerToken = "INSANITY"
local powerType = 13
local maxInsanity = 100
local isVfActive = false
local vfCurrentStackStartTime = 0
local vfStacks = 0
local vfDrainStacks = 0
--local isDispersionAvtive = false
local isVoiTActive = false
local isLiActive = false
local liStacks = 0
local displayLiStacks = nil

local isPiActive = false
local isLotvSelected = false
local isAsSelected = false
local isFotmSelected = false
local isLiSelected = false

local asTargets = {}

local insanityHexColor = "FFFFFF"
local castHexColor = "FFFFFF"
local asHexColor = "FFFFFF"

local isStmActive = false

local stmReportEnabled = 0
local maxArchivedReports = 4
local currentEncounterName = nil
local stmReport = {}

--local enablePatch = false
local vfBaseDrain = 6		-- Drain at the first stack of vf
local vfDrainFactor = 0.67
local vfThreshold = 65
local stmActiveMult = 2.0
local fotmActiveMult = 1.2
local piActiveMult = 1.25
local asInsanityGain = 3
local mbInsanityGain = 15
local svwInsanityGain = 25
local vtInsanityGain = 6
local mfInsanityGain = 3

local defaultOptions = {
	profile = {
		general = {
			posx = 0,
			posy = 0,
			width = 250,
			height = 22,
			alphaooc = 0.3,
			orientation = "HORIZONTAL",
		},
		bars = {
			texture = "Blizzard Raid Bar",
			insbarcolor = { 0.4, 0, 0.8, 1 },
			castbarcolor = { 1, 1, 1, 1 },
			asbarcolor = { 1, 0.31, 0.85, 1 },
		},
		background = {
			color = { 0.2, 0.2, 0.2, 0.5 },
		},
		border = {
			thickness = 1,
			color = { 0.0, 0.0, 0.0, 0.7 },
		},
		overlay = {
			thickness = 1,
			color = { 0, 0, 0, 1 },
		},
		text = {
			file = "Friz Quadrata TT",
			flag = "OUTLINE",
			insanitycolor = { 1, 1, 1, 1 },
			insanitysize = 12,
			castcolor = { 1, 1, 1, 1 },
			ascolor = { 1, 0.31, 0.85, 1 },
			vfstackcolor = { 1, 1, 1, 1 },
			vfstacksize = 12,
			vftimecolor = { 1, 1, 1, 1 },
			vftimesize = 12,
		},
		misc = {
			lotvthreshold = 65,
			lingeringinsanity = true,
			stmreportingenabled = 3,
			stmreportingmaxsaved = 3,
			--patch70105enable = false,
		},
		animations = {
			oocduration = 0.5,
			oocstartdelay = 1,
			bordercolor = { 0.7, 0.0, 1.0, 1.0 },
			borderduration = 1,
		},
		stmreportarchive = {
		},
	},
}

-- Print functions
addon.DebugEnabled = false
function addon:PrintDebug(str, ...)
	if self.DebugEnabled then
		print(format("|cFFFF0000" .. "Vilinka's Insanity Debug:|r " .. str, ...))
	end
end

function addon:Print(str, ...)
	print(format("|cFF9370DB" .. "Vilinka's Insanity:|r " .. str, ...))
end
-- Print functions

local function createBorder(parent, name, strata)
	local function CreateBorderTextures(border, layer, sublevel)
		local textures = {}
		-- Left --
		textures[1] = border:CreateTexture(nil, layer, nil, sublevel)
		textures[1]:SetSize(1, 0)
		textures[1]:SetPoint("TOPRIGHT", border, "TOPLEFT", 0, 1)
		textures[1]:SetPoint("BOTTOMRIGHT", border, "BOTTOMLEFT", 0, -1)
		textures[1]:SetColorTexture(1, 1, 1, 1)
		-- Right --
		textures[2] = border:CreateTexture(nil, layer, nil, sublevel)
		textures[2]:SetSize(1, 0)
		textures[2]:SetPoint("TOPLEFT", border, "TOPRIGHT", 0, 1)
		textures[2]:SetPoint("BOTTOMLEFT", border, "BOTTOMRIGHT", 0, -1)
		textures[2]:SetColorTexture(1, 1, 1, 1)
		-- Bottom --
		textures[3] = border:CreateTexture(nil, layer, nil, sublevel)
		textures[3]:SetSize(0, 1)
		textures[3]:SetPoint("TOPRIGHT", border, "BOTTOMLEFT")
		textures[3]:SetPoint("TOPLEFT", border, "BOTTOMRIGHT")
		textures[3]:SetColorTexture(1, 1, 1, 1)
		-- Top --
		textures[4] = border:CreateTexture(nil, layer, nil, sublevel)
		textures[4]:SetSize(0, 1)
		textures[4]:SetPoint("BOTTOMRIGHT", border, "TOPRIGHT")
		textures[4]:SetPoint("BOTTOMLEFT", border, "TOPLEFT")
		textures[4]:SetColorTexture(1, 1, 1, 1)
		
		textures.SetVertexColor = function(self, r, g, b, a)
			for i, texture in ipairs(self) do
				texture:SetVertexColor(r, g, b, a)
			end
		end
		
		textures.SetThickness = function(self, thicness)
			self[1]:SetPoint("TOPRIGHT", self[1]:GetParent(), "TOPLEFT", 0, thicness)
			self[1]:SetPoint("BOTTOMRIGHT", self[1]:GetParent(), "BOTTOMLEFT", 0, -thicness)
			self[1]:SetSize(thicness, 0)
			self[2]:SetPoint("TOPLEFT", self[2]:GetParent(), "TOPRIGHT", 0, thicness)
			self[2]:SetPoint("BOTTOMLEFT", self[2]:GetParent(), "BOTTOMRIGHT", 0, -thicness)
			self[2]:SetSize(thicness, 0)
			self[3]:SetSize(0, thicness)
			self[4]:SetSize(0, thicness)
		end
		
		return textures
	end
	local border = CreateFrame("Frame", name, parent)
	border:SetFrameStrata(strata)
	border:SetPoint("TOPLEFT", parent, "TOPLEFT")
	border:SetPoint("BOTTOMRIGHT", parent, "BOTTOMRIGHT")
	border.Textures = CreateBorderTextures(border, "BACKGROUND", -8)
	
	return border
end

local function createPredictionBarTexture(parent, layer, sublevel)
	local predictionBarTexture = parent:CreateTexture(nil, layer, nil, sublevel)
	
	predictionBarTexture.SetValue = function(self, currentValue, predictedValue)
		local value = currentValue + predictedValue
		if value > maxInsanity then
			predictedValue = maxInsanity - currentValue
		end
		local parent = self:GetParent()
		if parent:GetOrientation() == "HORIZONTAL" then
			local width = parent:GetWidth()
			local left = currentValue / maxInsanity
			local right = (currentValue + predictedValue) / maxInsanity
			local leftPosition = (currentValue / maxInsanity) * width
			local barWidth = (predictedValue / maxInsanity) * width
			if barWidth < 0.5 then
				self:Hide()
				return
			end
			self:ClearAllPoints()
			self:SetTexCoord(left, right, 0, 1)
			self:SetWidth(barWidth)
			self:SetPoint("TOPLEFT", leftPosition, 0)
			self:Show()
		else
			local height = parent:GetHeight()
			local bottom = currentValue / maxInsanity
			local top = (currentValue + predictedValue) / maxInsanity
			local bottomPosition = (currentValue / maxInsanity) * height
			local barHeight = (predictedValue / maxInsanity) * height
			if barHeight < 0.5 then
				self:Hide()
				return
			end
			self:ClearAllPoints()
			self:SetTexCoord(0, 1, bottom, top)
			self:SetHeight(barHeight)
			self:SetPoint("BOTTOMLEFT", 0, bottomPosition)
			self:Show()
		end
	end
	
	return predictionBarTexture
end

local insanityBar = CreateFrame("StatusBar", name .. "Bar", UIParent)
	
insanityBar.Background = CreateFrame("Frame", name .. "Background", insanityBar)
insanityBar.Background:SetFrameStrata("BACKGROUND")
insanityBar.Background:SetPoint("TOPLEFT", insanityBar, "TOPLEFT")
insanityBar.Background:SetPoint("BOTTOMRIGHT", insanityBar, "BOTTOMRIGHT")
insanityBar.Background.Texture = insanityBar.Background:CreateTexture(nil, "BACKGROUND", nil, -8)
insanityBar.Background.Texture:SetColorTexture(1, 1, 1, 1)
insanityBar.Background.Texture:SetPoint("TOPLEFT", insanityBar, "TOPLEFT")
insanityBar.Background.Texture:SetPoint("BOTTOMRIGHT", insanityBar, "BOTTOMRIGHT")

insanityBar.BorderNormal = createBorder(insanityBar, name .. "BorderNormal", "BACKGROUND")
insanityBar.BorderNormal:SetPoint("TOPLEFT", insanityBar, "TOPLEFT")
insanityBar.BorderNormal:SetPoint("BOTTOMRIGHT", insanityBar, "BOTTOMRIGHT")

insanityBar.BorderAnimated = createBorder(insanityBar, name .. "BorderAnimated", "MEDIUM")
insanityBar.BorderAnimated:SetPoint("TOPLEFT", insanityBar, "TOPLEFT")
insanityBar.BorderAnimated:SetPoint("BOTTOMRIGHT", insanityBar, "BOTTOMRIGHT")

insanityBar.CastPredictionBar = createPredictionBarTexture(insanityBar, "ARTWORK", 0)
insanityBar.ASPredictionBar = createPredictionBarTexture(insanityBar, "ARTWORK", 0)

insanityBar.EnterVoidformMark = CreateFrame("Frame", name .. "EnterVoidformMark", insanityBar)
insanityBar.EnterVoidformMark.Texture = insanityBar.EnterVoidformMark:CreateTexture(nil, "ARTWORK", nil, 1)
insanityBar.EnterVoidformMark.Texture:SetPoint("TOPLEFT", insanityBar.EnterVoidformMark, "TOPLEFT")
insanityBar.EnterVoidformMark.Texture:SetPoint("BOTTOMRIGHT", insanityBar.EnterVoidformMark, "BOTTOMRIGHT")
insanityBar.EnterVoidformMark.Texture:SetColorTexture(1, 1, 1, 1)

insanityBar.InsanityFontString = CreateFrame("Frame", name .. "InsanityFontStringFrame", insanityBar):CreateFontString(nil, "OVERLAY")
insanityBar.VoidformStackFontString = CreateFrame("Frame", name .. "VoidformStackFontStringFrame", insanityBar):CreateFontString(nil, "OVERLAY")
insanityBar.VoidformTimeFontString = CreateFrame("Frame", name .. "VoidformTimeFontStringFrame", insanityBar):CreateFontString(nil, "OVERLAY")

insanityBar.OOCFadeAnim = insanityBar:CreateAnimationGroup()
insanityBar.OOCFadeAnim:SetToFinalAlpha(true)
insanityBar.OOCFadeAnim.Alpha = insanityBar.OOCFadeAnim:CreateAnimation("Alpha")
insanityBar.OOCFadeAnim.Alpha:SetFromAlpha(1.0)
insanityBar.OOCFadeAnim.Alpha:SetSmoothing("NONE")

insanityBar.VoidformReadyAnim = insanityBar.BorderAnimated:CreateAnimationGroup()
insanityBar.VoidformReadyAnim:SetToFinalAlpha(true)
insanityBar.VoidformReadyAnim:SetLooping("REPEAT")
insanityBar.VoidformReadyAnim.AlphaIn = insanityBar.VoidformReadyAnim:CreateAnimation("Alpha")
insanityBar.VoidformReadyAnim.AlphaIn:SetSmoothing("OUT")
insanityBar.VoidformReadyAnim.AlphaIn:SetOrder(1)
insanityBar.VoidformReadyAnim.AlphaIn:SetFromAlpha(0.0)
insanityBar.VoidformReadyAnim.AlphaIn:SetToAlpha(1.0)
insanityBar.VoidformReadyAnim.AlphaOut = insanityBar.VoidformReadyAnim:CreateAnimation("Alpha")
insanityBar.VoidformReadyAnim.AlphaOut:SetSmoothing("IN")
insanityBar.VoidformReadyAnim.AlphaOut:SetOrder(2)
insanityBar.VoidformReadyAnim.AlphaOut:SetFromAlpha(1.0)
insanityBar.VoidformReadyAnim.AlphaOut:SetToAlpha(0.0)

insanityBar.VoidformEndAnim = insanityBar.BorderAnimated:CreateAnimationGroup()
insanityBar.VoidformEndAnim:SetToFinalAlpha(true)
insanityBar.VoidformEndAnim.Alpha = insanityBar.VoidformEndAnim:CreateAnimation("Alpha")
insanityBar.VoidformEndAnim.Alpha:SetSmoothing("NONE")
insanityBar.VoidformEndAnim.Alpha:SetFromAlpha(1.0)
insanityBar.VoidformEndAnim.Alpha:SetToAlpha(0.0)
insanityBar.VoidformEndAnim.Alpha:SetDuration(0.1)

insanityBar:SetScript("OnEvent", function(self, event, ...)
	self[event](self, ...)
end)

addon.InsanityBar = insanityBar

-- Conversion function from RGBA to HEX
-- Credit: https://gist.github.com/marceloCodget/3862929
-- passing a table like {1, 0.5, 0.8, 1}
local function rgbaToHex(rgba)
	local hexadecimal = ''
	
	-- Convert to wow supported format AARRGGBB
	local argb = { rgba[4]*255, rgba[1]*255, rgba[2]*255, rgba[3]*255 }

	for key, value in pairs(argb) do
		local hex = ''

		while(value > 0)do
			local index = math.fmod(value, 16) + 1
			value = math.floor(value / 16)
			hex = string.sub('0123456789ABCDEF', index, index) .. hex			
		end

		if(string.len(hex) == 0)then
			hex = '00'

		elseif(string.len(hex) == 1)then
			hex = '0' .. hex
		end

		hexadecimal = hexadecimal .. hex
	end

	return hexadecimal
end

local function ClearAsTargets(guid)
	if asTargets[guid] then
		asTargets[guid].asCount = nil
		asTargets[guid].lastUpdateTime = nil
		asTargets[guid] = nil
	end
end

local function GetVoidformTime(currentInsanity)
	if isVfActive then
		local insanity = currentInsanity
		local vfTime = 0
		if insanity > 0 then
			local currentTime = GetTime()
		
			-- Drain current stack
			local latency = select(4, GetNetStats()) / 1000
			local vfCurrentStackTimeLeft = 1.0 - (currentTime - vfCurrentStackStartTime) + latency
			local vfCurrentStackDrain = vfBaseDrain + ((vfDrainStacks - 1) * vfDrainFactor)
			if ((vfCurrentStackDrain * vfCurrentStackTimeLeft) > insanity) then
				vfCurrentStackTimeLeft = insanity / vfCurrentStackDrain
				insanity = 0
			else
				insanity = insanity - (vfCurrentStackDrain * vfCurrentStackTimeLeft)
			end
			
			--[[
				Sum d+((x-1)/2)*f = sum d+((x-1)*f'), Sum x=m to n => (1/2)*(-m+n+1)*(2*d+f*(m+n-2)) = d'
				m = start stack, n = end stack, d = initial drain, f' = drain factor, d' = insanity drained
				Solution of quadratic equation for n:
					n = (sqrt(4 d^2+8 d f' m-12 d f'+4 f'^2 m^2-12 f'^2 m+9 f'^2+8 f d')-2 d+f')/(2 f')
			]]--
			local m = vfDrainStacks + 1
			local tempSqrt = 4*vfBaseDrain*vfBaseDrain+8*vfBaseDrain*vfDrainFactor*m-12*vfBaseDrain*vfDrainFactor+4*vfDrainFactor*vfDrainFactor*m*m-12*vfDrainFactor*vfDrainFactor*m+9*vfDrainFactor*vfDrainFactor+8*vfDrainFactor*insanity
			local vfRemainingStacks = (sqrt(tempSqrt) - 2*vfBaseDrain + vfDrainFactor) / (2*vfDrainFactor)
			local vfRemainingStacksTime = vfRemainingStacks - vfDrainStacks
			vfTime = vfCurrentStackTimeLeft + vfRemainingStacksTime
		end
		
		return vfTime
	else
		return 0
	end
end

local function GetSpellInsanityGain(spellId)
	local fotmMult, piMult, stmMult = 1, 1, 1
	
	if isFotmSelected then
		fotmMult = fotmActiveMult
	end
	if isPiActive then
		piMult = piActiveMult
	end
	if isStmActive then
		stmMult = stmActiveMult
	end
	
	-- Mind Blast Id: 8092 --
	if spellId == 8092 then
		return mbInsanityGain * fotmMult * piMult * stmMult
	-- Mind Spike Id: 73510 --
	--elseif spellId == 73510 then
		--return 5 * fotmMult * piMult * stmMult	-- Patch: 7.1.5 Remove
	-- Shadow Word: Void Id: 205351 --
	elseif spellId == 205351 then
		return svwInsanityGain * piMult * stmMult
	-- Vampiric Touch Id: 34914 --
	elseif spellId == 34914 then
		return vtInsanityGain * piMult * stmMult
	-- Mind Flay Id: 15407 --
	elseif spellId == "Mind Flay" then
		return mfInsanityGain * fotmMult * piMult * stmMult
	-- Mind Sear Id: 48045 -
	--elseif spellId == "Mind Sear" then
		--return 4.5 * piMult * stmMult	-- Patch: 7.1.5 Remove
	else
		return 0
	end
end

local function CanEnterVoidform(currentInsanity)
	local insanityThreshold = 100
	if isLotvSelected then
		insanityThreshold = vfThreshold
	end
	if currentInsanity >= insanityThreshold then
		return true
	else
		return false
	end
end

local function printStmReport(report, fromArchive)
	if stmReportEnabled == 1 or stmReportEnabled == 3 or fromArchive == true then
		local totalduration = (report.totalduration == nil and 0 or report.totalduration)
		local vfduration = (report.vfduration == nil and 0 or report.vfduration)
		local vfstacks = (report.vfstacks == nil and 0 or report.vfstacks)
		local vfdrainstacks = (report.vfdrainstacks == nil and 0 or report.vfdrainstacks)
		addon:Print(format("STM report %s", date("%X %x", report.time)))
		if report.encounter ~= nil then
			print("Encounter: " .. report.encounter)
		end
		print(format("STM duration: %.1fs", totalduration))
		print(format("Voidform duration: %.1fs", vfduration))
		print(format("Voidform stacks: %d", vfstacks))
		print(format("Voidform drain stacks: %d", vfdrainstacks))
	end
end

function insanityBar:ADDON_LOADED(...)
	if name == ...	then
		self.db = LibStub("AceDB-3.0"):New("VilinkasInsanityDB", defaultOptions, true)
		self:Setup()
		
		self:ACTIVE_TALENT_GROUP_CHANGED()
		self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
	end
end

function insanityBar:ACTIVE_TALENT_GROUP_CHANGED()
	local playerSpec = GetSpecialization()
	if playerSpec == SPEC_PRIEST_SHADOW then
		--addon:PrintDebug("ACTIVE_TALENT_GROUP_CHANGED: " .. "Shadow spec is active.")
		self:RegisterEvent("PLAYER_TALENT_UPDATE")
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("ENCOUNTER_START")
		self:RegisterEvent("ENCOUNTER_END")
		self:RegisterUnitEvent("UNIT_MAXPOWER", "player")
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		
		self:SetScript("OnUpdate", VilinkasInsanityBar_OnUpdate)
		
		self:PLAYER_TALENT_UPDATE()
		
		self:Show()
		self:PLAYER_ENTERING_WORLD()
	else
		--addon:PrintDebug("ACTIVE_TALENT_GROUP_CHANGED: " .. "Shadow spec is not active.")
		self:UnregisterEvent("PLAYER_TALENT_UPDATE")
		self:UnregisterEvent("PLAYER_ENTERING_WORLD")
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
		self:UnregisterEvent("PLAYER_REGEN_DISABLED")
		self:UnregisterEvent("ENCOUNTER_START")
		self:UnregisterEvent("ENCOUNTER_END")
		self:UnregisterEvent("UNIT_MAXPOWER", "player")
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		
		self:SetScript("OnUpdate", nil)
		
		self:Hide()
	end
end

function insanityBar:PLAYER_TALENT_UPDATE()
	--addon:PrintDebug("PLAYER_TALENT_UPDATE")
	isFotmSelected = select(4, GetTalentInfo(1, 2, 1))
	isAsSelected = select(4, GetTalentInfo(5, 2, 1))
	if isAsSelected == false then
		asTargets = {}
	end
	isLotvSelected = select(4, GetTalentInfo(7, 1, 1))
	if isLotvSelected then
		self.EnterVoidformMark:Show()
	else
		self.EnterVoidformMark:Hide()
	end
	isLiSelected = select(4, GetTalentInfo(4, 1, 1))
		
	self:UNIT_MAXPOWER()
end

function insanityBar:PLAYER_ENTERING_WORLD()
	if not UnitAffectingCombat("player") then
		local oocAlpha = self.OOCFadeAnim.Alpha:GetToAlpha()
		self:SetAlpha(oocAlpha)
	end
	asTargets = {}
end

function insanityBar:PLAYER_REGEN_ENABLED()
	self.OOCFadeAnim:Play()
	asTargets = {}
	--addon:PrintDebug("Out of combat")
end

function insanityBar:PLAYER_REGEN_DISABLED()
	self.OOCFadeAnim:Stop()
	self:SetAlpha(1)
	--addon:PrintDebug("In combat")
end

function insanityBar:UNIT_MAXPOWER()
	maxInsanity = UnitPowerMax("player", powerToken)
	self:SetMinMaxValues(0, maxInsanity)
end

function insanityBar:ENCOUNTER_START(encounterID, encounterName)
	currentEncounterName = encounterName
end

function insanityBar:ENCOUNTER_END(encounterID, encounterName)
	currentEncounterName = nil
end

function insanityBar:COMBAT_LOG_EVENT_UNFILTERED(_, event, _, sourceGUID, _, _, _, destGUID, _, _, _, spellId)--, spellName)
	--local _, event, _, sourceGUID, sourceName, _, _, destGUID, destName, _, _, spellId, spellName = ...
	--addon:PrintDebug("COMBAT_LOG_EVENT: event: %s, spell id: %s, spell name: %s", event, spellId, spellName)
	if sourceGUID == UnitGUID("player") then
		-- Voidform --
		if spellId == 194249 then
			if event == "SPELL_AURA_APPLIED" then
				isVfActive = true
				vfCurrentStackStartTime = GetTime()
				vfDrainStacks = 1
				vfStacks = select(4, UnitBuff("player", "Voidform"))
				self.VoidformEndAnim:Stop()
				self.VoidformReadyAnim:Stop()
				self.BorderAnimated:SetAlpha(1.0)
				stmReport.vfduration = GetTime()
				--addon:PrintDebug("VF SPELL_AURA_APPLIED")
			elseif event == "SPELL_AURA_APPLIED_DOSE" then
				vfCurrentStackStartTime = GetTime()
				if not isVoiTActive then
					vfDrainStacks = vfDrainStacks + 1
					stmReport.vfdrainstacks = vfDrainStacks
				end
				vfStacks = vfStacks + 1
				stmReport.vfstacks = vfStacks
			elseif event == "SPELL_AURA_REMOVED" then
				isVfActive = false
				vfDrainStacks = 0
				vfStacks = 0
				vfCurrentStackStartTime = 0
				self.VoidformEndAnim:Play()
				--addon:PrintDebug("VF SPELL_AURA_REMOVED")
			end
		-- Lingering Insianity --
		elseif spellId == 197937 then
			if event == "SPELL_AURA_APPLIED" then
				isLiActive = true
				liStacks = select(4, UnitBuff("player", "Lingering Insanity"))
			elseif event == "SPELL_AURA_REMOVED_DOSE" then
				liStacks = liStacks - 2
			elseif event == "SPELL_AURA_REMOVED" then
				isLiActive = false
				liStacks = 0
			end
		-- Power Infusion --
		elseif spellId == 10060 then
			if event == "SPELL_AURA_APPLIED" then
				isPiActive = true
			elseif event == "SPELL_AURA_REMOVED" then
				isPiActive = false
			end
		-- Surrender to Madness --
		elseif spellId == 193223 then
			if event == "SPELL_AURA_APPLIED" then
				isStmActive = true
				stmReport.time = time()
				stmReport.totalduration = GetTime()
				stmReport.encounter = currentEncounterName
			elseif event == "SPELL_AURA_REMOVED" then
				isStmActive = false
				self:SaveSTMReport(stmReport)
				printStmReport(stmReport)
				stmReport = {}
			end
		-- Auspicious Spirits Spawn --
		elseif spellId == 147193 then
			if event == "SPELL_CAST_SUCCESS" then
				if isAsSelected then
					if not asTargets[destGUID] then
						asTargets[destGUID] = {}
						asTargets[destGUID].asCount = 0
					end
					asTargets[destGUID].asCount = asTargets[destGUID].asCount + 1
					asTargets[destGUID].lastUpdateTime = GetTime()
				end
			end
		-- Auspicious Spirits Despawn --
		elseif spellId == 148859 then
			if event == "SPELL_DAMAGE" then
				if isAsSelected then
					if asTargets[destGUID] then
						asTargets[destGUID].asCount = asTargets[destGUID].asCount - 1
						asTargets[destGUID].lastUpdateTime = GetTime()
						if asTargets[destGUID].asCount <= 0 then
							ClearAsTargets(guid)
						end
					end
				end
			end
		-- Dispersion --
		--[[elseif spellId == 47585 then
			if event == "SPELL_AURA_APPLIED" then
				isDispersionAvtive = true
			elseif event == "SPELL_AURA_REMOVED" then
				isDispersionAvtive = false
			end]]--
		-- Void Torrent --
		elseif spellId == 205065 then
			if event == "SPELL_AURA_APPLIED" then
				isVoiTActive = true
			elseif event == "SPELL_AURA_REMOVED" then
				isVoiTActive = false
			end
		end
	end
	if event == "UNIT_DIED" then
		if destGUID == UnitGUID("player") then
			asTargets = {}
			vfCurrentStackStartTime = 0
			vfDrainStacks = 0
			vfStacks = 0
			isVoiTActive = false
			isLiActive = false
			isPiActive = false
			self.VoidformReadyAnim:Stop()
			self.VoidformEndAnim:Play()
			if isStmActive == true then
				self:SaveSTMReport(stmReport)
				printStmReport(stmReport)
			end
			isVfActive = false
			isStmActive = false
			stmReport = {}
			encounterInProgress = false
		else
			if asTargets[destGUID] then
				ClearAsTargets(destGUID)
			end
		end
	end
end

function insanityBar:SetInsanityValue(insanity, castGainPrediction, asGainPrediction)
	self:SetValue(insanity)
	self.CastPredictionBar:SetValue(insanity, castGainPrediction)
	self.ASPredictionBar:SetValue(insanity + castGainPrediction, asGainPrediction)
	local castGainPredictionText
	local asGainPredictionText
	if self:GetOrientation() == "HORIZONTAL" then
		castGainPredictionText = castGainPrediction > 0 and "|c" .. castHexColor .. castGainPrediction .. "|r + " or ""
		asGainPredictionText = asGainPrediction > 0 and "|c" .. asHexColor .. asGainPrediction .. "|r + " or ""
		self.InsanityFontString:SetText(asGainPredictionText .. castGainPredictionText .. insanity)
	else
		castGainPredictionText = castGainPrediction > 0 and "\n+\n|c" .. castHexColor .. castGainPrediction .. "|r" or ""
		asGainPredictionText = asGainPrediction > 0 and "\n+\n|c" .. asHexColor .. asGainPrediction .. "|r" or ""
		self.InsanityFontString:SetText(format("%d", insanity) .. castGainPredictionText .. asGainPredictionText)
	end
	
	
end

function insanityBar:SetVfStackValue(stacks, haste)
	if self:GetOrientation() == "HORIZONTAL" then
		if haste == nil then
			self.VoidformStackFontString:SetText(format("%d", stacks))
		else
			self.VoidformStackFontString:SetText(format("%d (%d%%)", stacks, haste))
		end
	else
		if haste == nil then
			self.VoidformStackFontString:SetText(format("%d", stacks))
		else
			self.VoidformStackFontString:SetText(format("%d\n(%d%%)", stacks, haste))
		end
	end
end

function insanityBar:SetVfTimeValue(vfTime)
	self.VoidformTimeFontString:SetText(format("%.1f", vfTime))
end

function VilinkasInsanityBar_OnUpdate(self)
	local insanity = UnitPower("player", powerType)
	
	if not isVfActive then
		local canEnterVoidform = CanEnterVoidform(insanity)
		if canEnterVoidform and not self.canEnterVoidform then
			self.VoidformEndAnim:Stop()
			self.VoidformReadyAnim:Play()
			--addon:PrintDebug("Can enter VF")
		elseif not canEnterVoidform and self.canEnterVoidform then
			self.VoidformReadyAnim:Finish()
			--addon:PrintDebug("Can't enter VF")
		end
		self.canEnterVoidform = canEnterVoidform
	end
	
	local castSpellId = select(10, UnitCastingInfo("player"))
	local channelSpellId = select(1, UnitChannelInfo("player"))
	
	local castGainPrediction = 0
	
	if castSpellId then
		castGainPrediction = GetSpellInsanityGain(castSpellId)
	elseif channelSpellId then
		castGainPrediction = GetSpellInsanityGain(channelSpellId)
	end
	
	if not UnitAffectingCombat("player") then
		if castGainPrediction > 0 then
			self.OOCFadeAnim:Stop()
			self:SetAlpha(1)
			self.prevCastGainedInsanity = true
		else
			if self.prevCastGainedInsanity then
				self.prevCastGainedInsanity = false
				self.OOCFadeAnim:Play()
			end
		end
	end
	
	local currentTime = GetTime()
	
	local asGainPrediction = 0
	for guid in pairs(asTargets) do
		if asTargets[guid] ~= nil then
			if (currentTime - asTargets[guid].lastUpdateTime) > 10 then
				ClearAsTargets(guid)
			else
				asGainPrediction = asGainPrediction + asTargets[guid].asCount*asInsanityGain
			end
		end
	end
	
	self:SetInsanityValue(insanity, castGainPrediction, asGainPrediction)
	
	if isLiActive or isVfActive then
		local playerSpellHaste = UnitSpellHaste("player")
		if isVfActive then
			local vfTime = GetVoidformTime(insanity)
			self:SetVfStackValue(vfStacks, playerSpellHaste)
			self:SetVfTimeValue(vfTime)
		else
			if (displayLiStacks == true and isLiSelected == true) then -- Patch: 7.1.5
				self:SetVfStackValue(liStacks)
			else
				self.VoidformStackFontString:SetText("")
			end
			self.VoidformTimeFontString:SetText("")
		end
	else
		self.VoidformStackFontString:SetText("")
		self.VoidformTimeFontString:SetText("")
	end
end

--[[function insanityBar:TestSaveSTMReport(t, n)
	for i = 1,n do
		local testReport = {
			["time"] = t + i*100,
			["totalduration"] = i,
			["vfduration"] = i,
			["vfstacks"] = i,
			["vfdrainstacks"] = i,
			["encounter"] = nil,
		}
		self:SaveSTMReport(testReport)
	end
end]]--

function insanityBar:SaveSTMReport(report)
	if stmReportEnabled == 1 or stmReportEnabled == 0 then return end
	local profile = self.db.profile
	local reportarchive = profile.stmreportarchive
	
	local size = #reportarchive
	
	local currentTime = GetTime()
	
	if report.vfduration == nil then report.vfduration = 0 end
	
	report.totalduration = currentTime - report.totalduration
	report.vfduration = currentTime - report.vfduration
	
	table.insert(reportarchive, report)
	
	local d = size - maxArchivedReports
	if d >= 0 then
		for i = 1,(d+1) do
			table.remove(reportarchive, 1)
		end
	end
end

function insanityBar:ClearSavedSTMReports()
	local profile = self.db.profile
	local reportarchive = profile.stmreportarchive

	for i = 1,#reportarchive do
		table.remove(reportarchive)
	end
	
	addon:Print("STM report")
	print("Saved reports cleared!")
end

function insanityBar:PrintSavedSTMReports()
	local profile = self.db.profile
	local reportarchive = profile.stmreportarchive
	
	local size = #reportarchive
	
	if size == 0 then
		addon:Print("STM report")
		print("You currently have no saved STM reports.")
		return
	end
	
	print("------------------------")
	for _,v in pairs(reportarchive) do
		printStmReport(v, true)
		print("------------------------")
	end
end

function insanityBar:Setup()
	local LSM = LibStub("LibSharedMedia-3.0")
	local profile = self.db.profile
	
	local orientation = profile.general.orientation
	
	self:ClearAllPoints()
	self:SetPoint("CENTER", profile.general.posx, profile.general.posy)
	self:SetSize(profile.general.width, profile.general.height)
	
	-- Background --
	self.Background.Texture:SetVertexColor(profile.background.color[1], profile.background.color[2], profile.background.color[3], profile.background.color[4])
	
	-- Border --
	self.BorderNormal.Textures:SetVertexColor(profile.border.color[1], profile.border.color[2], profile.border.color[3], profile.border.color[4])
	self.BorderNormal.Textures:SetThickness(profile.border.thickness)
	self.BorderAnimated.Textures:SetVertexColor(profile.animations.bordercolor[1], profile.animations.bordercolor[2], profile.animations.bordercolor[3], profile.animations.bordercolor[4])
	self.BorderAnimated.Textures:SetThickness(profile.border.thickness)
	
	-- Bars --
	local bartexture = LSM:Fetch("statusbar", profile.bars.texture)
	
	-- Insanity Bar --
	self:SetStatusBarTexture(bartexture)
	self:SetStatusBarColor(profile.bars.insbarcolor[1], profile.bars.insbarcolor[2], profile.bars.insbarcolor[3], 1)
	self:SetOrientation(orientation)
	
	-- Cast Bar --
	self.CastPredictionBar:SetTexture(bartexture)
	self.CastPredictionBar:SetVertexColor(profile.bars.castbarcolor[1], profile.bars.castbarcolor[2], profile.bars.castbarcolor[3], 1)
	if orientation == "HORIZONTAL" then
		self.CastPredictionBar:SetHeight(profile.general.height)
		self.CastPredictionBar:SetRotation(0)
	else
		self.CastPredictionBar:SetWidth(profile.general.width)
		self.CastPredictionBar:SetRotation(90)
	end
	self.CastPredictionBar:Hide()
	
	-- AS Bar --
	self.ASPredictionBar:SetTexture(bartexture)
	self.ASPredictionBar:SetVertexColor(profile.bars.asbarcolor[1], profile.bars.asbarcolor[2], profile.bars.asbarcolor[3], 1)
	if orientation == "HORIZONTAL" then
		self.ASPredictionBar:SetHeight(profile.general.height)
		self.ASPredictionBar:SetRotation(0)
	else
		self.ASPredictionBar:SetWidth(profile.general.width)
		self.ASPredictionBar:SetRotation(90)
	end
	self.ASPredictionBar:Hide()
	
	-- Enter Voidform Mark --
	vfThreshold = profile.misc.lotvthreshold
	if orientation == "HORIZONTAL" then
		local width = profile.general.width
		local t = (profile.overlay.thickness / 2)
		local offsetX = vfThreshold / maxInsanity
		self.EnterVoidformMark:ClearAllPoints()
		self.EnterVoidformMark:SetPoint("TOPLEFT", self, "TOPLEFT", width * offsetX - t, 0)
		self.EnterVoidformMark:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -width * (1 - offsetX) + t, 0)
	else
		self.EnterVoidformMark:SetWidth(profile.general.width)
		self.EnterVoidformMark:SetHeight(profile.overlay.thickness)
		local markOffsetY = (self:GetHeight() / maxInsanity) * vfThreshold
		self.EnterVoidformMark:ClearAllPoints()
		self.EnterVoidformMark:SetPoint("CENTER", self, "BOTTOM", 0, markOffsetY)
	end
	self.EnterVoidformMark.Texture:SetVertexColor(profile.overlay.color[1], profile.overlay.color[2], profile.overlay.color[3], 1)
	
	-- Font --
	local file = LSM:Fetch("font", profile.text.file)
	local flag = profile.text.flag
	self.InsanityFontString:SetFont(file, profile.text.insanitysize, flag)
	self.InsanityFontString:SetTextColor(profile.text.insanitycolor[1], profile.text.insanitycolor[2], profile.text.insanitycolor[3], 1)
	self.VoidformTimeFontString:SetFont(file, profile.text.vftimesize, flag)
	self.VoidformTimeFontString:SetTextColor(profile.text.vftimecolor[1], profile.text.vftimecolor[2], profile.text.vftimecolor[3], 1)
	self.VoidformStackFontString:SetFont(file, profile.text.vfstacksize, flag)
	self.VoidformStackFontString:SetTextColor(profile.text.vfstackcolor[1], profile.text.vfstackcolor[2], profile.text.vfstackcolor[3], 1)
	castHexColor = rgbaToHex(profile.text.castcolor)
	asHexColor = rgbaToHex(profile.text.ascolor)
	if orientation == "HORIZONTAL" then
		self.InsanityFontString:ClearAllPoints()
		self.InsanityFontString:SetPoint("RIGHT", self, "RIGHT", -2, 0)
		self.VoidformTimeFontString:ClearAllPoints()
		self.VoidformTimeFontString:SetPoint("LEFT", self, "LEFT", 2, 0)
		self.VoidformStackFontString:ClearAllPoints()
		self.VoidformStackFontString:SetPoint("CENTER", self, "CENTER")
	else
		local halfHeight = profile.general.height / 2
		self.InsanityFontString:ClearAllPoints()
		self.InsanityFontString:SetPoint("TOPLEFT", self, "TOPLEFT", -5, -2)
		self.InsanityFontString:SetPoint("TOPRIGHT", self, "TOPRIGHT", 5, -2)
		self.InsanityFontString:SetJustifyH("CENTER")
		self.VoidformTimeFontString:ClearAllPoints()
		self.VoidformTimeFontString:SetPoint("BOTTOM", self, "BOTTOM", 0, 2)
		self.VoidformStackFontString:ClearAllPoints()
		self.VoidformStackFontString:SetPoint("TOP", self, "TOP", 0, -halfHeight + profile.text.vfstacksize)
	end
	
	-- Animations --
	self.OOCFadeAnim.Alpha:SetDuration(profile.animations.oocduration)
	self.OOCFadeAnim.Alpha:SetStartDelay(profile.animations.oocstartdelay)
	self.OOCFadeAnim.Alpha:SetToAlpha(profile.general.alphaooc)
	
	self.BorderAnimated:SetAlpha(0)
	self.VoidformReadyAnim.AlphaIn:SetDuration(profile.animations.borderduration / 2)
	self.VoidformReadyAnim.AlphaOut:SetDuration(profile.animations.borderduration / 2)
	
	-- Lingering Insanity --
	displayLiStacks = profile.misc.lingeringinsanity
	
	-- STM reporting --
	stmReportEnabled = profile.misc.stmreportingenabled
	maxArchivedReports	= profile.misc.stmreportingmaxsaved
	
	-- Patching --
	--[[enablePatch = profile.misc.patch70105enable
	if enablePatch == true then
	end]]--
end

insanityBar:RegisterEvent("ADDON_LOADED")

SLASH_VILINKASINSANITY1, SLASH_VILINKASINSANITY2 = "/vilinkasinsanity", "/vilins"

SlashCmdList["VILINKASINSANITY"] = function(message)
	local _, playerClass = UnitClass("player")
	local playerSpec = GetSpecialization()
	
	if playerClass == "PRIEST" then
		if playerSpec == SPEC_PRIEST_SHADOW then
			if message == "report" or message == "stm" then
				insanityBar:PrintSavedSTMReports()
			elseif message == "opt" then
				local loaded, reason = LoadAddOn("VilinkasInsanityOptions")

				if not loaded then
					addon:Print("Options module could not be loaded: " .. _G["ADDON_" .. reason])
					return
				end
				
				if UnitAffectingCombat("player") then
					addon:Print("options will open when you leave comat.")
					_G["VilinkasInsanityOptions"].showWhenOutOfCombat = true
				else
					_G["VilinkasInsanityOptions"]:Show()
				end
			else
				addon:Print("")
				print("Use |cFF9370DB/vilins opt|r to acces options menu.")
				print("Use |cFF9370DB/vilins stm|r display archived STM reports.")
			end
		else
			addon:Print("Current specialization is not supported.")
		end
	end
end
